#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import unittest
from typing import List

import mock
from freezegun import freeze_time
from mock import MagicMock

from neptune.internal.api_clients import HostedNeptuneBackendApiClient
from neptune.internal.api_clients.hosted_api_clients.hosted_alpha_leaderboard_api_client import \
    HostedAlphaLeaderboardApiClient
from neptune.internal.channels.channels import ChannelIdWithValues, ChannelNamespace, ChannelValue, ChannelType
from tests.neptune.new.backend_test_mixin import BackendTestMixin as AlphaBackendTestMixin

API_TOKEN = 'eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYWxwaGEuc3RhZ2UubmVwdHVuZS5haSIsImFwaV91cmwiOiJodHRwczovL2FscG' \
            'hhLnN0YWdlLm5lcHR1bmUuYWkiLCJhcGlfa2V5IjoiZDg5MGQ3Y2ItZGEzNi00MjRkLWJhNTQtZmVjZDJmYTdhOTQzIn0='
"""base64 decoded `API_TOKEN`
{
  "api_address": "https://alpha.stage.neptune.ai",
  "api_url": "https://alpha.stage.neptune.ai",
  "api_key": "d890d7cb-da36-424d-ba54-fecd2fa7a943"
}
"""


class TestAlphaIntegrationNeptuneBackend(unittest.TestCase, AlphaBackendTestMixin):
    @mock.patch('bravado.client.SwaggerClient.from_url')
    @mock.patch('neptune.internal.api_clients.hosted_api_clients.hosted_backend_api_client.NeptuneAuthenticator',
                new=MagicMock)
    @mock.patch('neptune.new.internal.backends.hosted_client.NeptuneAuthenticator', new=MagicMock)
    def setUp(self, swagger_client_factory) -> None:
        # pylint:disable=arguments-differ
        self._get_swagger_client_mock(swagger_client_factory)
        self.backend = HostedNeptuneBackendApiClient(API_TOKEN)
        self.leaderboard = HostedAlphaLeaderboardApiClient(self.backend)
        self.exp_mock = MagicMock(
            internal_id='00000000-0000-0000-0000-000000000000'
        )

    @freeze_time()
    def _test_send_channel_values(
            self, channel_y_elements: List[tuple], expected_operation: str, channel_type: ChannelType):
        # given prepared `ChannelIdWithValues`
        channel_id = 'channel_id'
        channel_name = 'channel_name'
        now_ms = int(time.time() * 1000)
        channel_with_values = ChannelIdWithValues(
            channel_id=channel_id,
            channel_name=channel_name,
            channel_type=channel_type.value,
            channel_namespace=ChannelNamespace.USER,
            channel_values=[
                ChannelValue(x=None, y={channel_y_key: channel_y_value}, ts=None)
                for channel_y_key, channel_y_value in channel_y_elements
            ]
        )

        # invoke send_channels_values
        self.leaderboard.send_channels_values(self.exp_mock, [channel_with_values])

        # expect `executeOperations` was called once with properly prepared kwargs
        expected_call_args = {
            'experimentId': '00000000-0000-0000-0000-000000000000',
            'operations': [{
                'path': f'logs/{channel_name}',
                expected_operation: {
                    'entries': [
                        {'value': channel_y_value, 'step': None, 'timestampMilliseconds': now_ms}
                        for _, channel_y_value in channel_y_elements
                    ]
                }
            }]
        }
        # pylint:disable=protected-access
        execute_operations = self.leaderboard.leaderboard_swagger_client.api.executeOperations
        self.assertEqual(len(execute_operations.call_args_list), 1)
        self.assertDictEqual(execute_operations.call_args_list[0][1], expected_call_args)

    def test_send_channels_text_values(self):
        channel_y_elements = [
            ('text_value', 'Line of text'),
            ('text_value', 'Another line of text'),
        ]
        self._test_send_channel_values(
            channel_y_elements, expected_operation='logStrings', channel_type=ChannelType.TEXT)

    def test_send_channels_numeric_values(self):
        channel_y_elements = [
            ('numeric_value', 42),
            ('numeric_value', 0.07),
        ]
        self._test_send_channel_values(
            channel_y_elements, expected_operation='logFloats', channel_type=ChannelType.NUMERIC)

    def test_send_channels_image_values(self):
        """TODO: implement in NPT-9207"""



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import hashlib
import pathlib

CHUNK_SIZE = 4096


# from https://stackoverflow.com/a/3431838
def md5(fname):
    hash_md5 = hashlib.md5()

    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(CHUNK_SIZE), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def append_non_relative_path(base_path: pathlib.Path, path_to_append: str) -> pathlib.Path:
    # By default when second path starts with '/', it replaces the path we're appending to
    relative_path = path_to_append[1:] if path_to_append.startswith('/') else path_to_append
    return base_path / relative_path



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, TYPE_CHECKING

from neptune.new.internal.backends.neptune_backend import NeptuneBackend

from neptune.new.internal.operation import Operation

if TYPE_CHECKING:
    from neptune.new.run import Run


class Attribute:

    def __init__(self, run: 'Run', path: List[str]):
        super().__init__()
        self._run = run
        self._path = path

    def __getattr__(self, attr):
        raise AttributeError("{} has no attribute {}.".format(type(self), attr))

    def _enqueue_operation(self, operation: Operation, wait: bool):
        # pylint: disable=protected-access
        self._run._op_processor.enqueue_operation(operation, wait)

    @property
    def _backend(self) -> NeptuneBackend:
        # pylint: disable=protected-access
        return self._run._backend

    @property
    def _run_id(self) -> str:
        # pylint: disable=protected-access
        return self._run._id



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import sqlite3 as sql
from pathlib import Path

from dataclasses import dataclass


class LocalFileHashStorage:
    @dataclass
    class LocalFileHash:
        file_path: str
        file_hash: str
        modification_date: str

    def __init__(self):
        db_path = Path.home() / ".neptune" / "files.db"
        os.makedirs(db_path.parent, exist_ok=True)

        self.session = sql.connect(str(db_path))
        self.cursor: sql.Cursor = self.session.cursor()
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS local_file_hashes"
            " (file_path text, file_hash text, modification_date text)"
        )
        self.session.commit()

    def insert(self, path: Path, computed_hash: str, modification_date: str):
        self.cursor.execute(
            f"INSERT INTO local_file_hashes"
            f" (file_path, file_hash, modification_date)"
            f" VALUES ('{str(path)}', '{computed_hash}', '{modification_date}')"
        )
        self.session.commit()

    def fetch_one(self, path: Path) -> 'LocalFileHash':
        found = [
            LocalFileHashStorage.LocalFileHash(*row)
            for row in self.cursor.execute(
                f"SELECT file_path, file_hash, modification_date"
                f" FROM local_file_hashes"
                f" WHERE file_path = '{str(path)}'"
            )
        ]

        return found[0] if found is not None and len(found) > 0 else None

    def update(self, path: Path, computed_hash: str, modification_date: str):
        self.cursor.execute(
            f"UPDATE local_file_hashes"
            f" SET file_hash='{computed_hash}', modification_date='{modification_date}'"
            f" WHERE file_path = '{str(path)}'"
        )
        self.session.commit()



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# pylint: disable=protected-access


from mock import MagicMock

from neptune.new.attributes.atoms.artifact import Artifact
from neptune.new.internal.operation import AssignArtifact
from neptune.new.types.atoms.artifact import Artifact as ArtifactVal
from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestArtifactHash(TestAttributeBase):
    def test_assign_type_error(self):
        values = ["foo", 10, None]
        for value in values:
            with self.assertRaises(Exception):
                Artifact(MagicMock(), MagicMock()).assign(value)

    def test_fetch(self):
        exp, path = self._create_run(), self._random_path()
        var = Artifact(exp, path)
        var._enqueue_operation(
            AssignArtifact(
                var._path,
                ArtifactVal("e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855").hash
            ),
            False)
        self.assertEqual(ArtifactVal("e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"), var.fetch())

    def test_fetch_hash(self):
        exp, path = self._create_run(), self._random_path()
        var = Artifact(exp, path)
        var._enqueue_operation(
            AssignArtifact(
                var._path,
                ArtifactVal("e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855").hash
            ),
            False)
        self.assertEqual("e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855", var.fetch_hash())



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mock import MagicMock

from tests.neptune.random_utils import a_string


def a_request():
    request = MagicMock()
    request.method = 'post'
    request.url = 'http://{}.com'.format(a_string())
    request.headers = {a_string(): a_string()}
    request.body = a_string()
    return request



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import unittest
import uuid
from datetime import datetime

from mock import Mock, patch

from neptune.new import ANONYMOUS, Run, get_last_run, get_project, init
from neptune.new.attributes.atoms import String
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import (
    MetadataInconsistency, NeptuneOfflineModeFetchException, NeptuneUninitializedException,
)
from neptune.new.internal.backends.api_model import (
    ApiRun,
    Attribute,
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
    IntAttribute,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.utils import IS_WINDOWS


@patch('neptune.new.internal.init_impl.HostedNeptuneBackend', NeptuneBackendMock)
class TestClient(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def test_incorrect_mode(self):
        with self.assertRaises(ValueError):
            init(mode='srtgj')

    def test_debug_mode(self):
        exp = init(mode='debug')
        exp["some/variable"] = 13
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_offline_mode(self):
        exp = init(mode='offline')
        exp["some/variable"] = 13
        with self.assertRaises(NeptuneOfflineModeFetchException):
            exp["some/variable"].fetch()
        self.assertIn(str(exp._id), os.listdir(".neptune/offline"))
        self.assertIn("data-1.log", os.listdir(".neptune/offline/{}".format(exp._id)))

    def test_sync_mode(self):
        exp = init(mode='sync')
        exp["some/variable"] = 13
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_async_mode(self):
        with init(mode='async', flush_period=0.5) as exp:
            exp["some/variable"] = 13
            with self.assertRaises(MetadataInconsistency):
                exp["some/variable"].fetch()
            exp.wait()
            self.assertEqual(13, exp["some/variable"].fetch())
            self.assertIn(str(exp._id), os.listdir(".neptune/async"))
            execution_dir = os.listdir(".neptune/async/{}".format(exp._id))[0]
            self.assertIn("data-1.log", os.listdir(".neptune/async/{}/{}".format(exp._id, execution_dir)))

    @patch("neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_run",
           new=lambda _, _id:
           ApiRun(uuid.UUID('12345678-1234-5678-1234-567812345678'), "SAN-94", "workspace", "sandbox", False))
    @patch("neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
           new=lambda _, _uuid: [Attribute("some/variable", AttributeType.INT)])
    @patch("neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_int_attribute",
           new=lambda _, _uuid, _path: IntAttribute(42))
    def test_read_only_mode(self):
        exp = init(mode='read-only', run="SAN-94")

        with self.assertLogs() as caplog:
            exp["some/variable"] = 13
            exp["some/other_variable"] = 11
            self.assertEqual(
                caplog.output,
                ['WARNING:neptune.new.internal.operation_processors.read_only_operation_processor:'
                 'Client in read-only mode, nothing will be saved to server.']
            )

        self.assertEqual(42, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    @patch("neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_run",
           new=lambda _, _id:
           ApiRun(uuid.UUID('12345678-1234-5678-1234-567812345678'), "SAN-94", "workspace", "sandbox", False))
    @patch("neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
           new=lambda _, _uuid: [Attribute("test", AttributeType.STRING)])
    def test_resume(self):
        with init(flush_period=0.5, run="SAN-94") as exp:
            self.assertEqual(exp._id, uuid.UUID('12345678-1234-5678-1234-567812345678'))
            self.assertIsInstance(exp.get_structure()["test"], String)

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.init_impl.os.path.isfile", new=lambda file: "." in file)
    @patch('neptune.new.internal.utils.glob', new=lambda path, recursive=False: [path.replace('*', 'file.txt')])
    @patch('neptune.new.internal.utils.os.path.abspath',
           new=lambda path: os.path.normpath("/home/user/main_dir/" + path))
    @patch('neptune.new.internal.utils.os.getcwd', new=lambda: "/home/user/main_dir")
    @unittest.skipIf(IS_WINDOWS, "Linux/Mac test")
    def test_entrypoint(self):
        exp = init(mode='debug')
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode='debug', source_files=[])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode='debug', source_files=["../*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main_dir/main.py")

        exp = init(mode='debug', source_files=["internal/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode='debug', source_files=["../other_dir/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "../main_dir/main.py")

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.is_ipython", new=lambda: True)
    def test_entrypoint_in_interactive_python(self):
        exp = init(mode='debug')
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode='debug', source_files=[])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode='debug', source_files=["../*"])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode='debug', source_files=["internal/*"])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.get_common_root", new=lambda _: None)
    @patch("neptune.new.internal.init_impl.os.path.isfile", new=lambda file: "." in file)
    @patch('neptune.new.internal.utils.glob', new=lambda path, recursive=False: [path.replace('*', 'file.txt')])
    @patch('neptune.new.internal.utils.os.path.abspath',
           new=lambda path: os.path.normpath("/home/user/main_dir/" + path))
    def test_entrypoint_without_common_root(self):
        exp = init(mode='debug', source_files=["../*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py")

        exp = init(mode='debug', source_files=["internal/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py")

    @patch("neptune.new.internal.get_project_impl.HostedNeptuneBackend")
    def test_get_table_as_pandas(self, backend_init_mock):
        # given
        backend_mock = Mock()
        backend_init_mock.return_value = backend_mock

        # and
        attributes = []
        now = datetime.now()
        attributes.append(AttributeWithProperties("run/state",
                                                  AttributeType.RUN_STATE,
                                                  Mock(value="idle")))
        attributes.append(AttributeWithProperties("float", AttributeType.FLOAT, Mock(value=12.5)))
        attributes.append(AttributeWithProperties("string", AttributeType.STRING, Mock(value="some text")))
        attributes.append(AttributeWithProperties("datetime", AttributeType.DATETIME, Mock(value=now)))
        attributes.append(AttributeWithProperties("float/series", AttributeType.FLOAT_SERIES, Mock(last=8.7)))
        attributes.append(AttributeWithProperties("string/series", AttributeType.STRING_SERIES, Mock(last="last text")))
        attributes.append(AttributeWithProperties("string/set", AttributeType.STRING_SET, Mock(values=["a", "b"])))
        attributes.append(AttributeWithProperties("git/ref",
                                                  AttributeType.GIT_REF,
                                                  Mock(commit=Mock(commitId="abcdef0123456789"))))
        attributes.append(AttributeWithProperties("file", AttributeType.FILE, None))
        attributes.append(AttributeWithProperties("file/set", AttributeType.FILE_SET, None))
        attributes.append(AttributeWithProperties("image/series", AttributeType.IMAGE_SERIES, None))

        # and
        empty_entry = LeaderboardEntry(uuid.uuid4(), [])
        filled_entry = LeaderboardEntry(uuid.uuid4(), attributes)
        backend_mock.get_leaderboard = Mock(return_value=[empty_entry, filled_entry])

        # when
        df = get_project().fetch_runs_table().to_pandas()

        # then
        self.assertEqual("idle", df['run/state'][1])
        self.assertEqual(12.5, df['float'][1])
        self.assertEqual("some text", df['string'][1])
        self.assertEqual(now, df['datetime'][1])
        self.assertEqual(8.7, df['float/series'][1])
        self.assertEqual("last text", df['string/series'][1])
        self.assertEqual("a,b", df['string/set'][1])
        self.assertEqual("abcdef0123456789", df['git/ref'][1])

        with self.assertRaises(KeyError):
            self.assertTrue(df['file'])
        with self.assertRaises(KeyError):
            self.assertTrue(df['file/set'])
        with self.assertRaises(KeyError):
            self.assertTrue(df['image/series'])

    @patch("neptune.new.internal.get_project_impl.HostedNeptuneBackend")
    def test_get_table_as_runs(self, backend_init_mock):
        # given
        backend_mock = Mock()
        backend_init_mock.return_value = backend_mock

        # and
        exp_id = uuid.uuid4()
        attributes = []
        now = datetime.now()
        attributes.append(AttributeWithProperties("run/state",
                                                  AttributeType.RUN_STATE,
                                                  Mock(value="idle")))
        attributes.append(AttributeWithProperties("float", AttributeType.FLOAT, Mock(value=12.5)))
        attributes.append(AttributeWithProperties("string", AttributeType.STRING, Mock(value="some text")))
        attributes.append(AttributeWithProperties("datetime", AttributeType.DATETIME, Mock(value=now)))
        attributes.append(AttributeWithProperties("float/series", AttributeType.FLOAT_SERIES, Mock(last=8.7)))
        attributes.append(AttributeWithProperties("string/series", AttributeType.STRING_SERIES, Mock(last="last text")))
        attributes.append(AttributeWithProperties("string/set", AttributeType.STRING_SET, Mock(values=["a", "b"])))
        attributes.append(AttributeWithProperties("git/ref",
                                                  AttributeType.GIT_REF,
                                                  Mock(commit=Mock(commitId="abcdef0123456789"))))
        attributes.append(AttributeWithProperties("file", AttributeType.FILE, None))
        attributes.append(AttributeWithProperties("file/set", AttributeType.FILE_SET, None))
        attributes.append(AttributeWithProperties("image/series", AttributeType.IMAGE_SERIES, None))

        # and
        backend_mock.get_leaderboard = Mock(return_value=[LeaderboardEntry(exp_id, attributes)])

        # when
        exp = get_project().fetch_runs_table().to_runs()[0]

        # then
        self.assertEqual("idle", exp['run/state'].get())
        self.assertEqual("idle", exp['run']['state'].get())
        self.assertEqual(12.5, exp['float'].get())
        self.assertEqual("some text", exp['string'].get())
        self.assertEqual(now, exp['datetime'].get())
        self.assertEqual(8.7, exp['float/series'].get())
        self.assertEqual("last text", exp['string/series'].get())
        self.assertEqual({"a", "b"}, exp['string/set'].get())
        self.assertEqual("abcdef0123456789", exp['git/ref'].get())

        with self.assertRaises(MetadataInconsistency):
            exp['file'].get()
        with self.assertRaises(MetadataInconsistency):
            exp['file/set'].get()
        with self.assertRaises(MetadataInconsistency):
            exp['image/series'].get()

        exp['file'].download("some_directory")
        backend_mock.download_file.assert_called_with(exp_id, ["file"], "some_directory")

        exp['file/set'].download("some_directory")
        backend_mock.download_file_set.assert_called_with(exp_id, ["file", "set"], "some_directory")

    def test_last_exp_is_raising_exception_when_non_initialized(self):
        # given uninitialized run
        Run.last_run = None

        # expect: raises NeptuneUninitializedException
        with self.assertRaises(NeptuneUninitializedException):
            get_last_run()

    def test_last_exp_is_the_latest_initialized(self):
        # given two initialized runs
        with init() as exp1, init() as exp2:
            # expect: `neptune.latest_run` to be the latest initialized one
            self.assertIsNot(exp1, get_last_run())
            self.assertIs(exp2, get_last_run())



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import logging
import os
import socket
import sys
import time
from functools import lru_cache, wraps
from typing import Optional, Dict, TYPE_CHECKING
from urllib.parse import urlparse, urljoin


import click
import requests
import urllib3
from urllib3.exceptions import NewConnectionError
from bravado.client import SwaggerClient
from bravado.exception import BravadoConnectionError, BravadoTimeoutError, HTTPForbidden, \
    HTTPServerError, HTTPUnauthorized, HTTPServiceUnavailable, HTTPRequestTimeout, \
    HTTPGatewayTimeout, HTTPBadGateway, HTTPClientError, HTTPTooManyRequests, HTTPError
from bravado.http_client import HttpClient
from bravado_core.formatter import SwaggerFormat
from packaging.version import Version
from requests import Session

from neptune.new.envs import NEPTUNE_RETRIES_TIMEOUT_ENV, NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE
from neptune.new.exceptions import SSLError, NeptuneConnectionLostException, \
    Unauthorized, Forbidden, CannotResolveHostname, UnsupportedClientVersion, ClientHttpError, \
    NeptuneFeaturesNotAvailableException
from neptune.new.internal.backends.api_model import ClientConfig
from neptune.new.internal.utils import replace_patch_version

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend


class OptionalFeatures:
    ARTIFACTS = "artifacts"


MAX_RETRY_TIME = 30
retries_timeout = int(os.getenv(NEPTUNE_RETRIES_TIMEOUT_ENV, "60"))


def with_api_exceptions_handler(func):

    def wrapper(*args, **kwargs):
        last_exception = None
        start_time = time.monotonic()
        for retry in itertools.count(0):
            if time.monotonic() - start_time > retries_timeout:
                break

            try:
                return func(*args, **kwargs)
            except requests.exceptions.SSLError as e:
                raise SSLError() from e
            except (BravadoConnectionError, BravadoTimeoutError,
                    requests.exceptions.ConnectionError, requests.exceptions.Timeout,
                    HTTPRequestTimeout, HTTPServiceUnavailable, HTTPGatewayTimeout, HTTPBadGateway,
                    HTTPTooManyRequests, HTTPServerError, NewConnectionError) as e:
                time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                last_exception = e
                continue
            except HTTPUnauthorized:
                raise Unauthorized()
            except HTTPForbidden:
                raise Forbidden()
            except HTTPClientError as e:
                raise ClientHttpError(e.status_code, e.response.text) from e
            except requests.exceptions.RequestException as e:
                if e.response is None:
                    raise
                status_code = e.response.status_code
                if status_code in (
                        HTTPRequestTimeout.status_code,
                        HTTPBadGateway.status_code,
                        HTTPServiceUnavailable.status_code,
                        HTTPGatewayTimeout.status_code,
                        HTTPTooManyRequests.status_code,
                        HTTPServerError.status_code):
                    time.sleep(min(2 ** min(10, retry), MAX_RETRY_TIME))
                    last_exception = e
                    continue
                elif status_code == HTTPUnauthorized.status_code:
                    raise Unauthorized()
                elif status_code == HTTPForbidden.status_code:
                    raise Forbidden()
                elif 400 <= status_code < 500:
                    raise ClientHttpError(status_code, e.response.text) from e
                else:
                    raise
        raise NeptuneConnectionLostException(last_exception) from last_exception

    return wrapper


@lru_cache(maxsize=None, typed=True)
def verify_host_resolution(url: str) -> None:
    host = urlparse(url).netloc.split(':')[0]
    try:
        socket.gethostbyname(host)
    except socket.gaierror:
        raise CannotResolveHostname(host)


uuid_format = SwaggerFormat(
    format='uuid',
    to_python=lambda x: x,
    to_wire=lambda x: x,
    validate=lambda x: None, description='')


@with_api_exceptions_handler
def create_swagger_client(url: str, http_client: HttpClient) -> SwaggerClient:
    return SwaggerClient.from_url(
        url,
        config=dict(
            validate_swagger_spec=False,
            validate_requests=False,
            validate_responses=False,
            formats=[uuid_format]
        ),
        http_client=http_client)


def verify_client_version(client_config: ClientConfig, version: Version):
    # Fix for failing in E2E when installing development version with pip and Github
    if str(version) == '0+unknown':
        return

    version_with_patch_0 = Version(replace_patch_version(str(version)))
    if client_config.min_compatible_version and client_config.min_compatible_version > version:
        raise UnsupportedClientVersion(version, min_version=client_config.min_compatible_version)
    if client_config.max_compatible_version and client_config.max_compatible_version < version_with_patch_0:
        raise UnsupportedClientVersion(version, max_version=client_config.max_compatible_version)
    if client_config.min_recommended_version and client_config.min_recommended_version > version:
        click.echo(
            "WARNING: There is a new version of neptune-client {} (installed: {}).".format(
                client_config.min_recommended_version, version),
            sys.stderr)


def update_session_proxies(session: Session, proxies: Optional[Dict[str, str]]):
    if proxies:
        try:
            session.proxies.update(proxies)
        except (TypeError, ValueError):
            raise ValueError("Wrong proxies format: {}".format(proxies))


def build_operation_url(base_api: str, operation_url: str) -> str:
    if '://' not in base_api:
        base_api = f'https://{base_api}'

    return urljoin(base=base_api, url=operation_url)


class MissingApiClient(SwaggerClient):
    """ catch-all class to gracefully handle calls to unavailable API """
    def __init__(self, backend: 'HostedNeptuneBackend'):  # pylint: disable=super-init-not-called
        self._backend = backend

    def __getattr__(self, item):
        raise NeptuneFeaturesNotAvailableException(features=self._backend.missing_features)


# https://stackoverflow.com/a/44776960
def cache(func):
    """
    Transform mutable dictionary into immutable before call to lru_cache
    """
    class HDict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    func = lru_cache(maxsize=None, typed=True)(func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        args = tuple([HDict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {k: HDict(v) if isinstance(v, dict) else v for k, v in kwargs.items()}
        return func(*args, **kwargs)

    wrapper.cache_clear = func.cache_clear
    return wrapper


def ssl_verify():
    if os.getenv(NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE):
        urllib3.disable_warnings()
        return False

    return True


def parse_validation_errors(error: HTTPError) -> Dict[str, str]:
    return {
        f"{error_description.get('errorCode').get('name')}": error_description.get('context', '')
        for validation_error in error.swagger_result.validationErrors
        for error_description in validation_error.get('errors')
    }



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import time
import hashlib
import unittest
import tempfile
from pathlib import Path

from mock import patch, Mock

from neptune.new.internal.artifacts.types import ArtifactFileData
from neptune.new.internal.artifacts.file_hasher import FileHasher


class TestFileHasher(unittest.TestCase):
    def setUp(self) -> None:
        self.temp = tempfile.TemporaryDirectory()

        with open(f'{self.temp.name}/test', 'wb') as handler:
            handler.write(b'\xde\xad\xbe\xef')

    def tearDown(self) -> None:
        self.temp.cleanup()

    def test_artifact_hash(self):
        # do not change this test case without coordinating with Artifact API's ArtifactHashComputer
        artifacts = [
            ArtifactFileData(
                file_path='to/file1',
                file_hash='c38444d2ccff1a7aab3d323fb6234e1b4f0a81ac',
                type="S3",
                size=5234,
                metadata={
                    'location': "s3://bucket/path/to/file1",
                    "last_modification": "2021-08-09 10:22:53"
                }
            ),
            ArtifactFileData(
                file_path='from/file2',
                file_hash='4347d0f8ba661234a8eadc005e2e1d1b646c9682',
                type="S3",
                metadata={
                    'location': "s3://bucket/path/to/file2",
                    "last_modification": "2021-08-09 10:32:12"
                }
            )
        ]

        expected_hash = "56e64245b1d4915ff27b306c8077cd4f9ce1b31233c690a93ebc38a1b737a9ea"
        self.assertEqual(expected_hash, FileHasher.get_artifact_hash(artifacts))
        self.assertEqual(expected_hash, FileHasher.get_artifact_hash(reversed(artifacts)))

    @patch('pathlib.Path.home')
    def test_local_file_hash(self, home):
        home.return_value = Path(self.temp.name)

        self.assertEqual(
            'd78f8bb992a56a597f6c7a1fb918bb78271367eb',
            FileHasher.get_local_file_hash(f'{self.temp.name}/test')
        )

    @patch('pathlib.Path.home')
    def test_local_file_hashed_only_once(self, home):
        home.return_value = Path(self.temp.name)
        hashlib.sha1 = Mock(side_effect=hashlib.sha1)

        hash1 = FileHasher.get_local_file_hash(f'{self.temp.name}/test')
        hash2 = FileHasher.get_local_file_hash(f'{self.temp.name}/test')

        self.assertEqual('d78f8bb992a56a597f6c7a1fb918bb78271367eb', hash1)
        self.assertEqual('d78f8bb992a56a597f6c7a1fb918bb78271367eb', hash2)
        self.assertEqual(1, hashlib.sha1.call_count)

    @patch('pathlib.Path.home')
    def test_local_file_hashed_update(self, home):
        home.return_value = Path(self.temp.name)
        hashlib.sha1 = Mock(side_effect=hashlib.sha1)

        hash1 = FileHasher.get_local_file_hash(f'{self.temp.name}/test')

        # Minimal change in modification time
        time.sleep(0.1)

        with open(f'{self.temp.name}/test', 'wb') as handler:
            handler.write(b'\x01\x02\x03\x04')

        hash2 = FileHasher.get_local_file_hash(f'{self.temp.name}/test')

        self.assertEqual('d78f8bb992a56a597f6c7a1fb918bb78271367eb', hash1)
        self.assertEqual('12dada1fff4d4787ade3333147202c3b443e376f', hash2)
        self.assertEqual(2, hashlib.sha1.call_count)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
import time
from typing import Optional, TypeVar, Generic, Union, Iterable

from neptune.new.internal.operation import Operation

from neptune.new.internal.utils import verify_type, is_collection

from neptune.new.types.series.series import Series as SeriesVal

from neptune.new.attributes.attribute import Attribute

Val = TypeVar('Val', bound=SeriesVal)
Data = TypeVar('Data')


class Series(Attribute, Generic[Val, Data]):

    def clear(self, wait: bool = False) -> None:
        self._clear_impl(wait)

    @abc.abstractmethod
    def _get_log_operation_from_value(self, value: Val, step: Optional[float], timestamp: float) -> Operation:
        pass

    # pylint: disable=unused-argument
    # pylint: disable=assignment-from-none
    def _get_config_operation_from_value(self, value: Val) -> Optional[Operation]:
        return None

    @abc.abstractmethod
    def _get_clear_operation(self) -> Operation:
        pass

    @abc.abstractmethod
    def _data_to_value(self, values: Iterable, **kwargs) -> Val:
        pass

    @abc.abstractmethod
    def _is_value_type(self, value) -> bool:
        pass

    def assign(self, value, wait: bool = False) -> None:
        if not self._is_value_type(value):
            value = self._data_to_value(value)
        clear_op = self._get_clear_operation()
        config_op = self._get_config_operation_from_value(value)
        with self._run.lock():
            if config_op:
                self._enqueue_operation(config_op, wait=False)
            if not value.values:
                self._enqueue_operation(clear_op, wait=wait)
            else:
                self._enqueue_operation(clear_op, wait=False)
                ts = time.time()
                self._enqueue_operation(self._get_log_operation_from_value(value, None, ts), wait=wait)

    def log(self,
            value: Union[Data, Iterable[Data]],
            step: Optional[float] = None,
            timestamp: Optional[float] = None,
            wait: bool = False,
            **kwargs) -> None:
        if is_collection(value):
            if step is not None and len(value) > 1:
                raise ValueError("Collection of values are not supported for explicitly defined 'step'.")
            value = self._data_to_value(value, **kwargs)
        else:
            value = self._data_to_value([value], **kwargs)

        if step is not None:
            verify_type("step", step, (float, int))
        if timestamp is not None:
            verify_type("timestamp", timestamp, (float, int))

        if not timestamp:
            timestamp = time.time()

        op = self._get_log_operation_from_value(value, step, timestamp)

        with self._run.lock():
            self._enqueue_operation(op, wait)

    def _clear_impl(self, wait: bool = False) -> None:
        op = self._get_clear_operation()
        with self._run.lock():
            self._enqueue_operation(op, wait)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Union, Optional, List
from urllib.parse import urlparse

from packaging.version import Version

from neptune.new import envs
from neptune.new.envs import CUSTOM_RUN_ID_ENV_NAME
from neptune.new.internal.utils import replace_patch_version
from neptune.new.internal.backends.api_model import Project, Workspace
from neptune.exceptions import STYLES


class NeptuneException(Exception):

    def __eq__(self, other):
        if type(other) is type(self):
            return super().__eq__(other) and str(self).__eq__(str(other))
        else:
            return False

    def __hash__(self):
        return hash((super().__hash__(), str(self)))


class NeptuneApiException(NeptuneException):
    pass


class MetadataInconsistency(NeptuneException):
    pass


class MissingFieldException(NeptuneException, AttributeError, KeyError):
    """Raised when get-like action is called on `Handler`, instead of on `Attribute`."""

    def __init__(self, field_path):
        message = """
{h1}
----MissingFieldException-------------------------------------------------------
{end}
Field "{field_path}" was not found.

There are two possible reasons:
    - There is a typo in a path. Double-check your code for typos.
    - You are fetching a field that other process created, but local representation is not synchronized.
    If you are sending metadata from multiple processes at the same time, synchronize the local representation before fetching values:
        {python}run.sync(){end}

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        self._msg = message.format(
            field_path=field_path,
            **STYLES
        )
        super().__init__(self._msg)

    def __str__(self):
        # required because of overriden `__str__` in `KeyError`
        return self._msg


class MalformedOperation(NeptuneException):
    pass


class FileNotFound(NeptuneException):
    def __init__(self, file: str):
        super().__init__("File not found: {}".format(file))


class FileUploadError(NeptuneException):
    def __init__(self, filename: str, msg: str):
        super().__init__("Cannot upload file {}: {}".format(filename, msg))


class FileSetUploadError(NeptuneException):
    def __init__(self, globs: List[str], msg: str):
        super().__init__("Cannot upload file set {}: {}".format(globs, msg))


class InternalClientError(NeptuneException):
    def __init__(self, msg: str):
        message = """
{h1}
----InternalClientError-----------------------------------------------------------------------
{end}
Neptune Client Library encountered an unexpected Internal Error:
{msg}

Please contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            msg=msg,
            **STYLES
        ))


class ClientHttpError(NeptuneException):
    def __init__(self, status, response):
        self.status = status
        self.response = response
        message = """
{h1}
----ClientHttpError-----------------------------------------------------------------------
{end}
Neptune server returned status {fail}{status}{end}.

Server response was:
{fail}{response}{end}

Verify the correctness of your call or contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            status=status,
            response=response,
            **STYLES
        ))


class ExceptionWithProjectsWorkspacesListing(NeptuneException):
    def __init__(self,
                 message: str,
                 available_projects: List[Project] = (),
                 available_workspaces: List[Workspace] = (),
                 **kwargs):
        available_projects_message = """
Did you mean any of these?
{projects}
"""

        available_workspaces_message = """
You can check all of your projects on the Projects page:
{workspaces_urls}
"""

        projects_formated_list = '\n'.join(
            map(lambda project: f'    - {project.workspace}/{project.name}', available_projects)
        )

        workspaces_formated_list = '\n'.join(
            map(lambda workspace: f'    - https://app.neptune.ai/{workspace.name}/-/projects', available_workspaces)
        )

        super().__init__(message.format(
            available_projects_message=available_projects_message.format(
                projects=projects_formated_list
            ) if available_projects else '',
            available_workspaces_message=available_workspaces_message.format(
                workspaces_urls=workspaces_formated_list
            ) if available_workspaces else '',
            **STYLES,
            **kwargs
        ))


class ProjectNotFound(ExceptionWithProjectsWorkspacesListing):
    def __init__(self,
                 project_id: str,
                 available_projects: List[Project] = (),
                 available_workspaces: List[Workspace] = ()):
        message = """
{h1}
----NeptuneProjectNotFoundException------------------------------------
{end}
We couldn’t find project {fail}"{project}"{end}.
{available_projects_message}{available_workspaces_message}
You may also want to check the following docs pages:
    - https://docs.neptune.ai/administration/projects
    - https://docs.neptune.ai/getting-started/hello-world#project

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message=message,
                         available_projects=available_projects,
                         available_workspaces=available_workspaces,
                         project=project_id)


class ProjectNameCollision(ExceptionWithProjectsWorkspacesListing):
    def __init__(self,
                 project_id: str,
                 available_projects: List[Project] = ()):
        message = """
{h1}
----NeptuneProjectNameCollisionException------------------------------------
{end}
Cannot resolve project {fail}"{project}"{end}.
{available_projects_message}
You may also want to check the following docs pages:
    - https://docs.neptune.ai/administration/projects
    - https://docs.neptune.ai/getting-started/hello-world#project

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message=message,
                         available_projects=available_projects,
                         project=project_id)


class NeptuneMissingProjectNameException(ExceptionWithProjectsWorkspacesListing):
    def __init__(self,
                 available_projects: List[Project] = (),
                 available_workspaces: List[Workspace] = ()):
        message = """
{h1}
----NeptuneMissingProjectNameException----------------------------------------
{end}
Neptune client couldn't find your project name.
{available_projects_message}{available_workspaces_message}
There are two options two add it:
    - specify it in your code 
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass it to {bold}neptune.init(){end} via {bold}project{end} argument:
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME'){end}

{h2}ENVIRONMENT VARIABLE{end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
       {bash}export {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

    {correct}Windows{end}
    In your CMD run:
       {bash}set {env_project}=WORKSPACE_NAME/PROJECT_NAME{end}

and skip the {bold}project{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/administration/projects
    - https://docs.neptune.ai/getting-started/hello-world#project

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message=message,
                         available_projects=available_projects,
                         available_workspaces=available_workspaces,
                         env_project=envs.PROJECT_ENV_NAME)


class RunNotFound(NeptuneException):

    def __init__(self, run_id: str) -> None:
        super().__init__("Run {} not found.".format(run_id))


class RunUUIDNotFound(NeptuneException):
    def __init__(self, run_id: str):
        super().__init__("Run with ID {} not found. Could be deleted.".format(run_id))


class InactiveRunException(NeptuneException):
    def __init__(self, short_id: str):
        message = """
{h1}
----InactiveRunException----------------------------------------
{end}
It seems you are trying to log (or fetch) metadata to a run that was stopped ({short_id}).
What should I do?
    - Resume the run to continue logging to it:
    https://docs.neptune.ai/how-to-guides/neptune-api/resume-run#how-to-resume-run
    - Don't invoke `stop()` on a run that you want to access. If you want to stop monitoring only, 
    you can resume a run in read-only mode:
    https://docs.neptune.ai/you-should-know/connection-modes#read-only
You may also want to check the following docs pages:
    - https://docs.neptune.ai/api-reference/run#stop
    - https://docs.neptune.ai/how-to-guides/neptune-api/resume-run#how-to-resume-run
    - https://docs.neptune.ai/you-should-know/connection-modes
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            short_id=short_id,
            **STYLES
        ))


class NeptuneMissingApiTokenException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneMissingApiTokenException-------------------------------------------
{end}
Neptune client couldn't find your API token.

You can get it here:
    - https://app.neptune.ai/get_my_api_token

There are two options to add it:
    - specify it in your code 
    - set an environment variable in your operating system.

{h2}CODE{end}
Pass the token to {bold}neptune.init(){end} via {bold}api_token{end} argument:
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}="YOUR_API_TOKEN"{end}

    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}="YOUR_API_TOKEN"{end}

and skip the {bold}api_token{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME'){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/getting-started/installation#authentication-neptune-api-token

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            env_api_token=envs.API_TOKEN_ENV_NAME,
            **STYLES
        ))


class NeptuneInvalidApiTokenException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneInvalidApiTokenException------------------------------------------------
{end}
Provided API token is invalid.
Make sure you copied and provided your API token correctly.

You can get it or check if it is correct here:
    - https://app.neptune.ai/get_my_api_token

There are two options to add it:
    - specify it in your code 
    - set as an environment variable in your operating system.

{h2}CODE{end}
Pass the token to {bold}neptune.init(){end} via {bold}api_token{end} argument:
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

{h2}ENVIRONMENT VARIABLE{end} {correct}(Recommended option){end}
or export or set an environment variable depending on your operating system: 

    {correct}Linux/Unix{end}
    In your terminal run:
        {bash}export {env_api_token}="YOUR_API_TOKEN"{end}

    {correct}Windows{end}
    In your CMD run:
        {bash}set {env_api_token}="YOUR_API_TOKEN"{end}

and skip the {bold}api_token{end} argument of {bold}neptune.init(){end}: 
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME'){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/getting-started/installation#authentication-neptune-api-token

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            env_api_token=envs.API_TOKEN_ENV_NAME,
            **STYLES
        ))


class CannotSynchronizeOfflineRunsWithoutProject(NeptuneException):
    def __init__(self):
        super().__init__("Cannot synchronize offline runs without a project.")


class NeedExistingRunForReadOnlyMode(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeedExistingRunForReadOnlyMode-----------------------------------------
{end}
Read-only mode can be used only with an existing run.

Parameter {python}run{end} of {python}neptune.init(){end} must be provided and reference
an existing run when using {python}mode="read-only"{end}.

You may also want to check the following docs pages:
    - https://docs.neptune.ai/you-should-know/connection-modes#read-only
    - https://docs.neptune.ai/api-reference/neptune#init

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneRunResumeAndCustomIdCollision(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneRunResumeAndCustomIdCollision-----------------------------------------
{end}
It's not possible to use {python}custom_run_id{end} while resuming a run.

Parameters {python}run{end} and {python}custom_run_id{end} of {python}neptune.init(){end} are mutually exclusive.
Make sure you have no {bash}{custom_id_env}{end} environment variable set
and no value is explicitly passed to `custom_run_id` argument when you are resuming a run.

You may also want to check the following docs pages:
    - https://docs.neptune.ai/api-reference/neptune#init

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            custom_id_env=CUSTOM_RUN_ID_ENV_NAME,
            **STYLES
        ))


class UnsupportedClientVersion(NeptuneException):
    def __init__(
            self,
            version: Union[Version, str],
            min_version: Optional[Union[Version, str]] = None,
            max_version: Optional[Union[Version, str]] = None):
        current_version = str(version)
        required_version = "==" + replace_patch_version(str(max_version)) if max_version else ">=" + str(min_version)
        message = """
{h1}
----UnsupportedClientVersion-------------------------------------------------------------
{end}
Your version of neptune-client ({current_version}) library is not supported by the Neptune server.

Please install neptune-client{required_version}

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            current_version=current_version,
            required_version=required_version,
            **STYLES
        ))


class CannotResolveHostname(NeptuneException):
    def __init__(self, host):
        message = """
{h1}
----CannotResolveHostname-----------------------------------------------------------------------
{end}
Neptune Client Library was not able to resolve hostname {underline}{host}{end}.

What should I do?
    - Check if your computer is connected to the internet.
    - Check if your computer should use any proxy to access internet.
      If so, you may want to use {python}proxies{end} parameter of {python}neptune.init(){end} function.
      See https://docs.neptune.ai/api-reference/neptune#init
      and https://requests.readthedocs.io/en/master/user/advanced/#proxies
    - Check Neptune services status: https://status.neptune.ai/

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            host=host,
            **STYLES
        ))


class SSLError(NeptuneException):
    def __init__(self):
        super().__init__(
            'SSL certificate validation failed. Set NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE '
            'environment variable to accept self-signed certificates.')


class NeptuneConnectionLostException(NeptuneException):
    def __init__(self, cause: Exception):
        self.cause = cause
        message = """
{h1}
----NeptuneConnectionLostException---------------------------------------------------------
{end}
A connection to the Neptune server was lost.
If you are using asynchronous (default) connection mode Neptune will continue to locally track your metadata and will continuously try to re-establish connection with Neptune servers.
If the connection is not re-established you can upload it later using Neptune Command Line Interface:
    {bash}neptune sync -p workspace_name/project_name{end}

What should I do?
    - Check if your computer is connected to the internet.
    - If your connection is unstable you can consider working using the offline mode:
        {python}run = neptune.init(mode="offline"){end}
        
You can read in detail how it works and how to upload your data on the following doc pages:
    - https://docs.neptune.ai/you-should-know/connection-modes#offline
    - https://docs.neptune.ai/you-should-know/connection-modes#uploading-offline-data
    
You may also want to check the following docs pages:
    - https://docs.neptune.ai/you-should-know/connection-modes#connectivity-issues
    - https://docs.neptune.ai/you-should-know/connection-modes
    
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class InternalServerError(NeptuneApiException):
    def __init__(self, response):
        message = """
{h1}
----InternalServerError-----------------------------------------------------------------------
{end}
Neptune Client Library encountered an unexpected Internal Server Error.

Server response was:
{fail}{response}{end}

Please try again later or contact Neptune support.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            response=response,
            **STYLES
        ))


class Unauthorized(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Unauthorized-----------------------------------------------------------------------
{end}
You have no permission to access given resource.
    
    - Verify your API token is correct.
      See: https://app.neptune.ai/get_my_api_token
      
    - Verify if your the provided project name is correct.
      The correct project name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose
          
   - Ask your organization administrator to grant you necessary privileges to the project

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class Forbidden(NeptuneApiException):
    def __init__(self):
        message = """
{h1}
----Forbidden-----------------------------------------------------------------------
{end}
You have no permission to access given resource.
    
    - Verify your API token is correct.
      See: https://app.neptune.ai/get_my_api_token
      
    - Verify if your the provided project name is correct.
      The correct project name should look like this {correct}WORKSPACE/PROJECT_NAME{end}.
      It has two parts:
          - {correct}WORKSPACE{end}: which can be your username or your organization name
          - {correct}PROJECT_NAME{end}: which is the actual project name you chose
          
   - Ask your organization administrator to grant you necessary privileges to the project

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneOfflineModeFetchException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneOfflineModeFetchException---------------------------------------------------
{end}
It seems you are trying to fetch data from the server, while working in an offline mode.
You need to work in non-offline connection mode to fetch data from the server. 

You can set connection mode when creating a new run:
    {python}run = neptune.init(mode="async"){end}
    
You may also want to check the following docs pages:
    - https://docs.neptune.ai/you-should-know/connection-modes
    
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class OperationNotSupported(NeptuneException):
    def __init__(self, message: str):
        super().__init__(f'Operation not supported: {message}')


class NeptuneLegacyProjectException(NeptuneException):
    def __init__(self, project: str):
        message = """
{h1}
----NeptuneLegacyProjectException---------------------------------------------------------
{end}
Your project "{project}" has not been migrated to the new structure yet.
Unfortunately neptune.new Python API is incompatible with projects using old structure,
please use legacy neptune Python API.
Don't worry - we are working hard on migrating all the projects and you will be able to use the neptune.new API soon. 

You can find documentation for legacy neptune Python API here:
    - https://docs-legacy.neptune.ai/index.html
    
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            project=project,
            **STYLES
        ))


class NeptuneUninitializedException(NeptuneException):
    def __init__(self):
        message = """
{h1}     
----NeptuneUninitializedException----------------------------------------------------
{end}
You must initialize neptune-client before you access `get_last_run`.

Looks like you forgot to add:
    {python}neptune.init(project='WORKSPACE_NAME/PROJECT_NAME', api_token='YOUR_API_TOKEN'){end}

before you ran:
    {python}neptune.get_last_run(){end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/api-reference/neptune#get_last_run

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneIntegrationNotInstalledException(NeptuneException):
    def __init__(self, framework):
        message = """
{h1}     
----NeptuneIntegrationNotInstalledException-----------------------------------------
{end}
Looks like integration neptune-{framework} wasn't installed.
To install run:
    {bash}pip install neptune-{framework}{end}
Or:
    {bash}pip install neptune-client[{framework}]{end}

You may also want to check the following docs pages:
    - https://docs.neptune.ai/integrations-and-supported-tools/intro
    
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(
            framework=framework,
            **STYLES
        ))


class NeptuneStorageLimitException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneStorageLimitException---------------------------------------------------------------------------------------
{end}
You exceeded storage limit for workspace. It's not possible to upload new data, but you can still fetch and delete data.
If you are using asynchronous (default) connection mode Neptune automatically switched to an offline mode
and your data is being stored safely on the disk. You can upload it later using Neptune Command Line Interface:
    {bash}neptune sync -p project_name{end}
What should I do?
    - Go to your projects and remove runs or model metadata you don't need
    - ... or update your subscription plan here: https://app.neptune.ai/-/subscription
You may also want to check the following docs pages:
    - https://docs.neptune.ai/advanced-user-guides/connection-modes
{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class FetchAttributeNotFoundException(MetadataInconsistency):
    def __init__(self, attribute_path: str):
        message = """
{h1}
----MetadataInconsistency----------------------------------------------------------------------
{end}
Field {python}{attribute_path}{end} was not found.

Remember that in the asynchronous (default) connection mode data is synchronized
with the Neptune servers in the background and may have not reached
it yet before it's fetched. Before fetching the data you can force
wait for all the requests sent by invoking:

    {python}run.wait(){end}
    
Remember that each use of {python}wait{end} introduces a delay in code execution.

You may also want to check the following docs pages:
    - https://docs.neptune.ai/you-should-know/connection-modes

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help.html
"""
        super().__init__(message.format(
            attribute_path=attribute_path,
            **STYLES
        ))


class ArtifactNotFoundException(MetadataInconsistency):
    def __init__(self, artifact_hash: str):
        message = """
{h1}
----MetadataInconsistency----------------------------------------------------------------------
{end}
Artifact with hash {python}{artifact_hash}{end} was not found.

Remember that in the asynchronous (default) connection mode data is synchronized
with the Neptune servers in the background and may have not reached
it yet before it's fetched. Before fetching the data you can force
wait for all the requests sent by invoking:

    {python}run.wait(){end}

Remember that each use of {python}wait{end} introduces a delay in code execution.

You may also want to check the following docs pages:
    - https://docs.neptune.ai/you-should-know/connection-modes

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help.html
"""
        super().__init__(message.format(
            artifact_hash=artifact_hash,
            **STYLES
        ))


class PlotlyIncompatibilityException(Exception):
    def __init__(self, matplotlib_version, plotly_version):
        super().__init__(
            "Unable to convert plotly figure to matplotlib format. "
            "Your matplotlib ({}) and plotlib ({}) versions are not compatible. "
            "See https://stackoverflow.com/q/63120058 for details. "
            "Downgrade matplotlib to version 3.2 or use as_image to log static chart.".format(
                matplotlib_version,
                plotly_version))


class NeptunePossibleLegacyUsageException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptunePossibleLegacyUsageException----------------------------------------------------------------
{end}
It seems you are trying to use legacy API, but imported the new one.

Simply update your import statement to:
    {python}import neptune{end}

You may want to check the Legacy API docs:
    - https://docs-legacy.neptune.ai

If you want to update your code with the new API we prepared a handy migration guide:
    - https://docs.neptune.ai/migration-guide

You can read more about neptune.new in the release blog post:
    - https://neptune.ai/blog/neptune-new

You may also want to check the following docs pages:
    - https://docs-legacy.neptune.ai/getting-started/integrate-neptune-into-your-codebase.html

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneLegacyIncompatibilityException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneLegacyIncompatibilityException----------------------------------------
{end}
It seems you are passing the legacy Experiment object, when a Run object is expected.

What can I do?
    - Updating your code to the new Python API requires few changes, but to help you with this process we prepared a handy migration guide:
    https://docs.neptune.ai/migration-guide
    - You can read more about neptune.new in the release blog post:
    https://neptune.ai/blog/neptune-new

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneUnhandledArtifactSchemeException(NeptuneException):
    def __init__(self, path: str):
        scheme = urlparse(path).scheme
        message = """
{h1}
----NeptuneUnhandledArtifactProtocolException------------------------------------
{end}
You have used a Neptune Artifact to track a file with scheme unhandled by this client ({scheme}).
Problematic path: {path}

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(scheme=scheme, path=path, **STYLES))


class NeptuneUnhandledArtifactTypeException(NeptuneException):
    def __init__(self, type_str: str):
        message = """
{h1}
----NeptuneUnhandledArtifactTypeException----------------------------------------
{end}
A Neptune Artifact you're listing is tracking a file type unhandled by this client ({type_str}).

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(type_str=type_str, **STYLES))


class NeptuneLocalStorageAccessException(NeptuneException):
    def __init__(self, path, expected_description):
        message = """
{h1}
----NeptuneLocalStorageAccessException-------------------------------------
{end}
Neptune had problem processing "{path}", it expects it to be {expected_description}.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(path=path, expected_description=expected_description, **STYLES))


class NeptuneRemoteStorageCredentialsException(NeptuneException):
    def __init__(self):
        message = """
{h1}
----NeptuneRemoteStorageCredentialsException-------------------------------------
{end}
Neptune could not find suitable credentials for remote storage of a Neptune Artifact you're listing.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(**STYLES))


class NeptuneRemoteStorageAccessException(NeptuneException):
    def __init__(self, location: str):
        message = """
{h1}
----NeptuneRemoteStorageAccessException------------------------------------------
{end}
Neptune could not access an object ({location}) from remote storage of a Neptune Artifact you're listing.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(location=location, **STYLES))


class ArtifactUploadingError(NeptuneException):
    def __init__(self, msg: str):
        super().__init__("Cannot upload artifact: {}".format(msg))


class NeptuneUnsupportedArtifactFunctionalityException(NeptuneException):
    def __init__(self, functionality_info: str):
        message = """
{h1}
----NeptuneUnsupportedArtifactFunctionality-------------------------------------
{end}
It seems you are using Neptune Artifacts functionality that is currently not supported.

{functionality_info}

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(functionality_info=functionality_info, **STYLES))


class NeptuneEmptyLocationException(NeptuneException):
    def __init__(self, location: str, namespace: str):
        message = """
{h1}
----NeptuneEmptyLocationException----------------------------------------------
{end}
Neptune could not find files in the requested location ({location}) during creation of an Artifact in "{namespace}".

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        super().__init__(message.format(location=location, namespace=namespace, **STYLES))


class NeptuneFeaturesNotAvailableException(NeptuneException):
    def __init__(self, features: List[str]):
        message = """
{h1}
----NeptuneFeaturesNotAvailableException----------------------------------------------
{end}
Following features are not yet supported by the Neptune instance you are using:
{features_desc}

An update of the Neptune instance is required in order to use them. Please contact your local Neptune administrator
or the Neptune support directly (support@neptune.ai) about the upcoming updates.

{correct}Need help?{end}-> https://docs.neptune.ai/getting-started/getting-help
"""
        features_desc = "\n".join(f"    - {f}" for f in features)
        self.message = message.format(features_desc=features_desc, **STYLES)
        super().__init__(message)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.new.internal.backends.utils import build_operation_url


class TestNeptuneBackendMock(unittest.TestCase):
    def test_building_operation_url(self):
        urls = {
            build_operation_url(
                'https://app.stage.neptune.ai',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai/',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'https://app.stage.neptune.ai/',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai',
                '/api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai/',
                'api/leaderboard/v1/attributes/download'
            ),
            build_operation_url(
                'app.stage.neptune.ai/',
                '/api/leaderboard/v1/attributes/download'
            ),
        }
        self.assertEqual({'https://app.stage.neptune.ai/api/leaderboard/v1/attributes/download'}, urls)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pathlib
import typing

from neptune.new.attributes.atoms.atom import Atom
from neptune.new.internal.artifacts.types import ArtifactDriver, ArtifactDriversMap, ArtifactFileData
from neptune.new.internal.backends.utils import OptionalFeatures
from neptune.new.internal.operation import AssignArtifact, TrackFilesToArtifact
from neptune.new.types.atoms.artifact import Artifact as ArtifactVal


class Artifact(Atom):
    def _check_feature(self):
        self._run._backend.verify_feature_available(OptionalFeatures.ARTIFACTS)  # pylint: disable=protected-access

    def assign(self, value: ArtifactVal, wait: bool = False):
        self._check_feature()
        # this function should be used only with ArtifactVal
        if not isinstance(value, ArtifactVal):
            raise TypeError("Value of unsupported type {}".format(type(value)))

        with self._run.lock():
            self._enqueue_operation(AssignArtifact(self._path, value.hash), wait)

    def fetch(self) -> ArtifactVal:
        self._check_feature()
        return ArtifactVal(self.fetch_hash())

    def fetch_hash(self) -> str:
        self._check_feature()
        val = self._backend.get_artifact_attribute(self._run_id, self._path)
        return val.hash

    def fetch_files_list(self) -> typing.List[ArtifactFileData]:
        self._check_feature()
        artifact_hash = self.fetch_hash()
        return self._backend.list_artifact_files(
            self._run._project_id,  # pylint: disable=protected-access
            artifact_hash
        )

    def download(self, destination: str = None):
        self._check_feature()
        for file_definition in self.fetch_files_list():
            driver: typing.Type[ArtifactDriver] = ArtifactDriversMap.match_type(file_definition.type)
            file_destination = pathlib.Path(destination or '.') / pathlib.Path(file_definition.file_path)
            file_destination.parent.mkdir(parents=True, exist_ok=True)
            driver.download_file(file_destination, file_definition)

    def track_files(
            self,
            path: str,
            destination: str = None,
            wait: bool = False
    ):
        self._check_feature()
        with self._run.lock():
            self._enqueue_operation(
                TrackFilesToArtifact(self._path,
                                     self._run._project_id,  # pylint: disable=protected-access
                                     [(path, destination)]),
                wait
            )



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
        # self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].pop('tag2_to_remove')
        # self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].pop('tag4_remove_non_existing')
* NEPTUNE_API_TOKEN
* NEPTUNE_PROJECT
"""
import sys
from datetime import datetime

from PIL import Image

import neptune.new as neptune
from alpha_integration_dev.common_client_code import ClientFeatures
from neptune.new.attributes.constants import (
    ARTIFACT_ATTRIBUTE_SPACE,
    LOG_ATTRIBUTE_SPACE,
    PROPERTIES_ATTRIBUTE_SPACE,
    SOURCE_CODE_FILES_ATTRIBUTE_PATH,
    SYSTEM_TAGS_ATTRIBUTE_PATH,
)
from neptune.new.exceptions import MissingFieldException
from neptune.new.types import File


class NewClientFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        self.exp = neptune.init(
            source_files='alpha_integration_dev/*.py'
        )

        # download sources
        self.exp.sync()
        with self.with_check_if_file_appears('files.zip'):
            self.exp[SOURCE_CODE_FILES_ATTRIBUTE_PATH].download()


    def modify_tags(self):
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].add('tag1')
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].add(['tag2_to_remove', 'tag3'])
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].remove('tag2_to_remove')
        self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].remove('tag4_remove_non_existing')
        # del self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH]  # TODO: NPT-9222

        self.exp.sync()
        assert self.exp[SYSTEM_TAGS_ATTRIBUTE_PATH].fetch() == {'tag1', 'tag3'}

    def modify_properties(self):
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop'] = 'some text'
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_number'] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['nested/prop'] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_to_del'] = 42
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_list'] = [1, 2, 3]
        with open(self.text_file_path, mode='r') as f:
            self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_IO'] = f
        self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_datetime'] = datetime.now()
        self.exp.sync()
        del self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_to_del']

        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop'].fetch() == 'some text'
        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_number'].fetch() == 42
        assert self.exp[PROPERTIES_ATTRIBUTE_SPACE]['nested/prop'].fetch() == 42
        prop_to_del_absent = False
        try:
            self.exp[PROPERTIES_ATTRIBUTE_SPACE]['prop_to_del'].fetch()
        except MissingFieldException:
            prop_to_del_absent = True
        assert prop_to_del_absent

    def log_std(self):
        print('stdout text1')
        print('stdout text2')
        print('stderr text1', file=sys.stderr)
        print('stderr text2', file=sys.stderr)

    def log_series(self):
        # floats
        self.exp[LOG_ATTRIBUTE_SPACE]['m1'].log(1)
        self.exp[LOG_ATTRIBUTE_SPACE]['m1'].log(2)
        self.exp[LOG_ATTRIBUTE_SPACE]['m1'].log(3)
        self.exp[LOG_ATTRIBUTE_SPACE]['m1'].log(2)
        self.exp[LOG_ATTRIBUTE_SPACE]['nested']['m1'].log(1)

        # texts
        self.exp[LOG_ATTRIBUTE_SPACE]['m2'].log('a')
        self.exp[LOG_ATTRIBUTE_SPACE]['m2'].log('b')
        self.exp[LOG_ATTRIBUTE_SPACE]['m2'].log('c')

        # images
        im_frame = Image.open(self.img_path)
        g_img = File.as_image(im_frame)
        self.exp[LOG_ATTRIBUTE_SPACE]['g_img'].log(g_img)

    def handle_files_and_images(self):
        # image
        im_frame = Image.open(self.img_path)
        g_img = File.as_image(im_frame)
        self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned image'] = g_img
        self.exp.wait()
        with self.with_check_if_file_appears('assigned image.png'):
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned image'].download()
        with self.with_check_if_file_appears('custom_dest.png'):
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned image'].download('custom_dest.png')

        self.exp[ARTIFACT_ATTRIBUTE_SPACE]['logged image'].log(g_img)

        with open(self.img_path, mode='r') as f:
            # self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned image stream'] = f
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]['logged image stream'].log(f)

        # artifact
        text_file = neptune.types.File(self.text_file_path)
        self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned file'] = text_file
        # self.exp[ARTIFACT_ATTRIBUTE_SPACE]['logged file'].log(text_file)  # wrong type
        with open(self.text_file_path, mode='r') as f:
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]['assigned file stream'] = f
            self.exp[ARTIFACT_ATTRIBUTE_SPACE]['logged file stream'].log(f)

    def handle_directories(self):
        pass

    def finalize(self):
        return


if __name__ == '__main__':
    NewClientFeatures().run()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from neptune.internal.hardware.constants import BYTES_IN_ONE_GB
from neptune.internal.hardware.gauges.gauge import Gauge
from neptune.internal.hardware.gpu.gpu_monitor import GPUMonitor


class GpuUsageGauge(Gauge):
    def __init__(self, card_index):
        self.card_index = card_index
        self.__gpu_monitor = GPUMonitor()

    def name(self):
        return str(self.card_index)

    def value(self):
        return self.__gpu_monitor.get_card_usage_percent(self.card_index)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.card_index == other.card_index

    def __repr__(self):
        return str(u'GpuUsageGauge')


class GpuMemoryGauge(Gauge):
    def __init__(self, card_index):
        self.card_index = card_index
        self.__gpu_monitor = GPUMonitor()

    def name(self):
        return str(self.card_index)

    def value(self):
        return self.__gpu_monitor.get_card_used_memory_in_bytes(self.card_index) / float(BYTES_IN_ONE_GB)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.card_index == other.card_index

    def __repr__(self):
        return str(u'GpuMemoryGauge')



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest
from unittest.mock import MagicMock

from neptune.new.attributes import create_attribute_from_type
from neptune.new.attributes.attribute import Attribute
from neptune.new.internal.backends.api_model import AttributeType


class TestAttributeUtils(unittest.TestCase):
    def test_attribute_type_to_atom(self):
        # Expect all AttributeTypes are reflected in `attribute_type_to_atom`...
        # ... and this reflection is class based on `Attribute`
        self.assertTrue(all(isinstance(create_attribute_from_type(attr_type, MagicMock(), ""), Attribute)
                            for attr_type in AttributeType))



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os

from mock import MagicMock

from neptune.new.internal.operation import UploadFileSet, DeleteFiles
from neptune.new.attributes.file_set import FileSet

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestFileSet(TestAttributeBase):

    def test_assign(self):
        globs = ["path1", "dir/", "glob/*.py"]
        expected = [os.path.abspath(glob) for glob in globs]

        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = FileSet(exp, path)
        var.assign(globs, wait=wait)
        processor.enqueue_operation.assert_called_once_with(UploadFileSet(path, expected, reset=True), wait)

    def test_upload_files(self):
        globs = ["path1", "dir/", "glob/*.py"]
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = FileSet(exp, path)
        var.upload_files(globs, wait=wait)
        processor.enqueue_operation.assert_called_once_with(
            UploadFileSet(path, [os.path.abspath(glob) for glob in globs], reset=False), wait)

    def test_delete_files(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = FileSet(exp, path)
        var.delete_files(["path1", "dir/"], wait=wait)
        processor.enqueue_operation.assert_called_once_with(DeleteFiles(path, {"path1", "dir/"}), wait)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

import uuid

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.backends.operations_preprocessor import OperationsPreprocessor
from neptune.new.internal.operation import (
    AddStrings,
    AssignFloat,
    AssignString,
    ClearFloatLog,
    ClearImageLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    LogFloats,
    LogImages,
    LogStrings,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFileSet
)

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase

FLog = LogFloats.ValueType
SLog = LogStrings.ValueType
ILog = LogImages.ValueType


class TestOperationsPreprocessor(TestAttributeBase):

    def test_delete_attribute(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process([
            DeleteAttribute(["a"]),
            AssignFloat(["a"], 1),
            DeleteAttribute(["a"]),

            AssignString(["b"], "2"),
            DeleteAttribute(["b"]),

            AssignString(["c"], "2"),
            AssignString(["c"], "3"),
            DeleteAttribute(["c"]),

            DeleteAttribute(["d"]),
            DeleteAttribute(["d"]),

            DeleteAttribute(["e"]),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            DeleteAttribute(["a"]),

            AssignString(["b"], "2"),
            DeleteAttribute(["b"]),

            AssignString(["c"], "3"),
            DeleteAttribute(["c"]),

            DeleteAttribute(["d"]),

            DeleteAttribute(["e"]),
        ])
        self.assertEqual(processor.get_errors(), [])

    def test_assign(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process([
            AssignFloat(["a"], 1),
            DeleteAttribute(["a"]),
            AssignString(["a"], "111"),

            DeleteAttribute(["b"]),
            AssignFloat(["b"], 2),

            AssignFloat(["c"], 3),
            AssignString(["c"], "333"),

            AssignString(["d"], "44"),

            AssignFloat(["e"], 5),
            AssignFloat(["e"], 10),
            AssignFloat(["e"], 33),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            AssignFloat(["a"], 1),
            DeleteAttribute(["a"]),
            AssignString(["a"], "111"),

            DeleteAttribute(["b"]),
            AssignFloat(["b"], 2),

            AssignFloat(["c"], 3),

            AssignString(["d"], "44"),

            AssignFloat(["e"], 33),
        ])
        self.assertEqual(processor.get_errors(), [
            MetadataInconsistency("Cannot perform AssignString operation on c: Attribute is not a String")
        ])

    def test_series(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process([
            LogFloats(["a"], [FLog(1, 2, 3)]),
            ConfigFloatSeries(["a"], min=7, max=70, unit="%"),
            DeleteAttribute(["a"]),
            LogStrings(["a"], [SLog("111", 3, 4)]),

            DeleteAttribute(["b"]),
            LogStrings(["b"], [SLog("222", None, 6)]),

            LogFloats(["c"], [FLog(1, 2, 3)]),
            LogFloats(["c"], [FLog(10, 20, 30), FLog(100, 200, 300)]),

            LogStrings(["d"], [SLog("4", 111, 222)]),

            ClearFloatLog(["e"]),

            LogImages(["f"], [ILog("1", 2, 3)]),
            LogImages(["f"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
            LogImages(["f"], [ILog("1", 2, 3)]),
            LogImages(["f"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
            ClearImageLog(["f"]),
            LogImages(["f"], [ILog("3", 20, 30), FLog("4", 200, 300)]),
            LogImages(["f"], [ILog("5", 2, 3)]),
            LogImages(["f"], [ILog("8", 20, 30), FLog("1000", 200, 300)]),

            LogImages(["g"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
            ClearImageLog(["g"]),

            AssignString(["h"], "44"),
            LogFloats(["h"], [FLog(10, 20, 30), FLog(100, 200, 300)]),

            LogFloats(["i"], [FLog(1, 2, 3)]),
            ConfigFloatSeries(["i"], min=7, max=70, unit="%"),
            ClearFloatLog(["i"]),
            LogFloats(["i"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            LogFloats(["a"], [FLog(1, 2, 3)]),
            DeleteAttribute(["a"]),
            LogStrings(["a"], [FLog("111", 3, 4)]),

            DeleteAttribute(["b"]),
            LogStrings(["b"], [SLog("222", None, 6)]),

            LogFloats(["c"], [FLog(1, 2, 3), FLog(10, 20, 30), FLog(100, 200, 300)]),

            LogStrings(["d"], [SLog("4", 111, 222)]),

            ClearFloatLog(["e"]),

            ClearImageLog(["f"]),
            LogImages(["f"], [ILog("3", 20, 30), FLog("4", 200, 300), ILog("5", 2, 3),
                              ILog("8", 20, 30), FLog("1000", 200, 300)]),

            ClearImageLog(["g"]),

            AssignString(["h"], "44"),

            ClearFloatLog(["i"]),
            LogFloats(["i"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
            ConfigFloatSeries(["i"], min=7, max=70, unit="%"),
        ])
        self.assertEqual(processor.get_errors(), [
            MetadataInconsistency("Cannot perform LogFloats operation on h: Attribute is not a Float Series")
        ])

    def test_sets(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process([
            AddStrings(["a"], {"xx", "y", "abc"}),
            DeleteAttribute(["a"]),
            AddStrings(["a"], {"hhh", "gij"}),

            DeleteAttribute(["b"]),
            RemoveStrings(["b"], {"abc", "defgh"}),

            AddStrings(["c"], {"hhh", "gij"}),
            RemoveStrings(["c"], {"abc", "defgh"}),
            AddStrings(["c"], {"qqq"}),

            ClearStringSet(["d"]),

            RemoveStrings(["e"], {"abc", "defgh"}),
            AddStrings(["e"], {"qqq"}),
            ClearStringSet(["e"]),

            AddStrings(["f"], {"hhh", "gij"}),
            RemoveStrings(["f"], {"abc", "defgh"}),
            AddStrings(["f"], {"qqq"}),
            ClearStringSet(["f"]),
            AddStrings(["f"], {"xx", "y", "abc"}),
            RemoveStrings(["f"], {"abc", "defgh"}),

            AssignString(["h"], "44"),
            RemoveStrings(["h"], {""}),

            AssignFloat(["i"], 5),
            AddStrings(["i"], {""}),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            AddStrings(["a"], {"xx", "y", "abc"}),
            DeleteAttribute(["a"]),
            AddStrings(["a"], {"hhh", "gij"}),

            DeleteAttribute(["b"]),
            RemoveStrings(["b"], {"abc", "defgh"}),

            AddStrings(["c"], {"hhh", "gij"}),
            RemoveStrings(["c"], {"abc", "defgh"}),
            AddStrings(["c"], {"qqq"}),

            ClearStringSet(["d"]),

            ClearStringSet(["e"]),

            ClearStringSet(["f"]),
            AddStrings(["f"], {"xx", "y", "abc"}),
            RemoveStrings(["f"], {"abc", "defgh"}),

            AssignString(["h"], "44"),
            AssignFloat(["i"], 5),
        ])
        self.assertEqual(processor.get_errors(), [
            MetadataInconsistency("Cannot perform RemoveStrings operation on h: Attribute is not a String Set"),
            MetadataInconsistency("Cannot perform AddStrings operation on i: Attribute is not a String Set")
        ])

    def test_file_set(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process([
            UploadFileSet(["a"], ["xx", "y", "abc"], reset=False),
            DeleteAttribute(["a"]),
            UploadFileSet(["a"], ["hhh", "gij"], reset=False),

            DeleteAttribute(["b"]),
            UploadFileSet(["b"], ["abc", "defgh"], reset=True),

            UploadFileSet(["c"], ["hhh", "gij"], reset=False),
            UploadFileSet(["c"], ["abc", "defgh"], reset=True),
            UploadFileSet(["c"], ["qqq"], reset=False),

            UploadFileSet(["d"], ["hhh", "gij"], reset=False),

            AssignFloat(["e"], 5),
            UploadFileSet(["e"], [""], reset=False),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            UploadFileSet(["a"], ["xx", "y", "abc"], reset=False),
            DeleteAttribute(["a"]),
            UploadFileSet(["a"], ["hhh", "gij"], reset=False),

            DeleteAttribute(["b"]),
            UploadFileSet(["b"], ["abc", "defgh"], reset=True),

            UploadFileSet(["c"], ["abc", "defgh"], reset=True),
            UploadFileSet(["c"], ["qqq"], reset=False),

            UploadFileSet(["d"], ["hhh", "gij"], reset=False),

            AssignFloat(["e"], 5),
        ])
        self.assertEqual(processor.get_errors(), [
            MetadataInconsistency("Cannot perform UploadFileSet operation on e: Attribute is not a File Set")
        ])

    def test_artifacts(self):
        # given
        processor = OperationsPreprocessor()
        project_uuid = uuid.uuid4()

        # when
        processor.process([
            TrackFilesToArtifact(["a"], project_uuid, [('dir1/', None)]),
            DeleteAttribute(["a"]),

            TrackFilesToArtifact(["b"], project_uuid, [('dir1/', None)]),
            TrackFilesToArtifact(["b"], project_uuid, [('dir2/dir3/', 'dir2/')]),
            TrackFilesToArtifact(["b"], project_uuid, [('dir4/dir5/', 'dir4/')]),

            AssignFloat(["c"], 5),
            TrackFilesToArtifact(["c"], project_uuid, [('dir1/', None)]),

            TrackFilesToArtifact(["d"], project_uuid, [('dir2/dir3/', 'dir2/')]),
            TrackFilesToArtifact(["d"], project_uuid, [('dir4/', None)]),

            TrackFilesToArtifact(["e"], project_uuid, [('dir1/', None)]),
            TrackFilesToArtifact(["e"], project_uuid, [('dir2/dir3/', 'dir2/')]),

            TrackFilesToArtifact(["f"], project_uuid, [('dir1/', None)]),
            TrackFilesToArtifact(["f"], project_uuid, [('dir2/dir3/', 'dir2/')]),
            TrackFilesToArtifact(["f"], project_uuid, [('dir4/', None)]),
        ])

        # then
        self.assertEqual(processor.get_operations(), [
            TrackFilesToArtifact(["a"], project_uuid, [('dir1/', None)]),
            DeleteAttribute(["a"]),

            TrackFilesToArtifact(
                ["b"],
                project_uuid,
                [('dir1/', None), ('dir2/dir3/', 'dir2/'), ('dir4/dir5/', 'dir4/')]
            ),

            AssignFloat(["c"], 5),

            TrackFilesToArtifact(["d"], project_uuid, [('dir2/dir3/', 'dir2/'), ('dir4/', None)]),

            TrackFilesToArtifact(["e"], project_uuid, [('dir1/', None), ('dir2/dir3/', 'dir2/')]),

            TrackFilesToArtifact(
                ["f"],
                project_uuid,
                [('dir1/', None), ('dir2/dir3/', 'dir2/'), ('dir4/', None)]
            ),
        ])
        self.assertEqual(processor.get_errors(), [
            MetadataInconsistency(
                "Cannot perform TrackFilesToArtifact operation on c: Attribute is not a Artifact"
            ),
        ])



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.new.internal.artifacts.types import ArtifactMetadataSerializer


class TestArtifactMetadataSerializer(unittest.TestCase):
    def test_simple(self):
        metadata = {
            'location': "s3://bucket/path/to/file",
            'last_modification': '2021-08-09 09:41:53',
            'file_size': '18'
        }

        serialized = ArtifactMetadataSerializer.serialize(metadata)

        self.assertListEqual(
            [
                {'key': 'file_size', 'value': '18'},
                {'key': 'last_modification', 'value': '2021-08-09 09:41:53'},
                {'key': 'location', 'value': 's3://bucket/path/to/file'}
            ],
            serialized
        )

        deserialized = ArtifactMetadataSerializer.deserialize(serialized)

        self.assertDictEqual(metadata, deserialized)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import pathlib
import shutil
import tempfile
import unittest
from pathlib import Path

from neptune.new.exceptions import NeptuneLocalStorageAccessException, NeptuneUnsupportedArtifactFunctionalityException
from neptune.new.internal.artifacts.drivers.local import LocalArtifactDriver
from neptune.new.internal.artifacts.types import ArtifactDriversMap, ArtifactFileData, ArtifactFileType
from tests.neptune.new.internal.artifacts.utils import md5


class TestLocalArtifactDrivers(unittest.TestCase):
    test_dir = None

    def setUp(self):
        self.test_sources_dir = Path(str(tempfile.mktemp()))
        self.test_dir = Path(str(tempfile.mktemp()))
        test_source_data = Path(__file__).parents[5] / 'data' / 'local_artifact_drivers_data'
        test_data = self.test_dir / 'data'

        # copy source data to temp dir (to prevent e.g. inter-fs symlinks)
        shutil.copytree(test_source_data, self.test_sources_dir)

        # create files to track
        shutil.copytree(self.test_sources_dir / 'files_to_track', test_data)

        # symbolic and hard link files
        # `link_to` is new in python 3.8
        # (test_source_data / 'file_to_link.txt').link_to(test_data / 'hardlinked_file.txt')
        os.link(
            src=str(self.test_sources_dir / 'file_to_link.txt'),
            dst=str(test_data / 'hardlinked_file.txt')
        )
        (test_data / 'symlinked_file.txt').symlink_to(self.test_sources_dir / 'file_to_link.txt')

        # symlink dir - content of this file won't be discovered
        (test_data / 'symlinked_dir').symlink_to(self.test_sources_dir / 'dir_to_link', target_is_directory=True)

    def tearDown(self) -> None:
        # clean tmp directories
        shutil.rmtree(self.test_dir, ignore_errors=True)
        shutil.rmtree(self.test_sources_dir, ignore_errors=True)

    def test_match_by_path(self):
        self.assertEqual(
            ArtifactDriversMap.match_path('file:///path/to/'),
            LocalArtifactDriver
        )
        self.assertEqual(
            ArtifactDriversMap.match_path('/path/to/'),
            LocalArtifactDriver
        )

    def test_match_by_type(self):
        self.assertEqual(
            ArtifactDriversMap.match_type('Local'),
            LocalArtifactDriver
        )

    def test_file_download(self):
        path = (self.test_dir / 'data/file1.txt').as_posix()
        artifact_file = ArtifactFileData(
            file_path='data/file1.txt',
            file_hash='??',
            type='??',
            metadata={
                "file_path": f"file://{path}"
            }
        )

        with tempfile.TemporaryDirectory() as temporary:
            downloaded_file = Path(temporary) / 'downloaded_file.ext'

            LocalArtifactDriver.download_file(
                destination=downloaded_file,
                file_definition=artifact_file
            )

            self.assertTrue(Path(downloaded_file).is_symlink())
            self.assertEqual('ad62f265e5b1a2dc51f531e44e748aa0', md5(downloaded_file))

    def test_non_existing_file_download(self):
        path = '/wrong/path'
        artifact_file = ArtifactFileData(
            file_path=path,
            file_hash='??',
            type='??',
            metadata={
                'file_path': path
            }
        )

        with self.assertRaises(NeptuneLocalStorageAccessException), tempfile.TemporaryDirectory() as temporary:
            local_destination = Path(temporary)
            LocalArtifactDriver.download_file(
                destination=local_destination,
                file_definition=artifact_file
            )

    def test_single_retrieval(self):
        files = LocalArtifactDriver.get_tracked_files(str(self.test_dir / 'data/file1.txt'))

        self.assertEqual(1, len(files))
        self.assertIsInstance(files[0], ArtifactFileData)
        self.assertEqual(ArtifactFileType.LOCAL.value, files[0].type)
        self.assertEqual('d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384', files[0].file_hash)
        self.assertEqual('file1.txt', files[0].file_path)
        self.assertEqual(21, files[0].size)
        self.assertEqual(
            {'file_path', 'last_modified'},
            files[0].metadata.keys()
        )
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata['file_path']
        )
        self.assertIsInstance(
            files[0].metadata['last_modified'],
            str
        )

    def test_multiple_retrieval(self):
        files = LocalArtifactDriver.get_tracked_files(str(self.test_dir / 'data'))
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(4, len(files))

        self.assertEqual('file1.txt', files[0].file_path)
        self.assertEqual('d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384', files[0].file_hash)
        self.assertEqual(21, files[0].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata['file_path'])

        self.assertEqual('hardlinked_file.txt', files[1].file_path)
        self.assertEqual('da3e6ddfa171e1ab5564609caa1dbbea9871886e', files[1].file_hash)
        self.assertEqual(44, files[1].size)
        self.assertEqual(f"file://{(self.test_dir.resolve() / 'data/hardlinked_file.txt').as_posix()}",
                         files[1].metadata['file_path'])

        self.assertEqual('sub_dir/file_in_subdir.txt', files[2].file_path)
        self.assertEqual('98181b1a4c880a462fcfa96b92c84b8e945ac335', files[2].file_hash)
        self.assertEqual(24, files[2].size)
        self.assertEqual(f"file://{(self.test_dir.resolve() / 'data/sub_dir/file_in_subdir.txt').as_posix()}",
                         files[2].metadata['file_path'])

        self.assertEqual('symlinked_file.txt', files[3].file_path)
        self.assertEqual('da3e6ddfa171e1ab5564609caa1dbbea9871886e', files[3].file_hash)
        self.assertEqual(44, files[3].size)
        self.assertEqual(f"file://{(self.test_sources_dir.resolve() / 'file_to_link.txt').as_posix()}",
                         files[3].metadata['file_path'])

    def test_multiple_retrieval_prefix(self):
        files = LocalArtifactDriver.get_tracked_files((self.test_dir / 'data').as_posix(), 'my/custom_path')
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(4, len(files))

        self.assertEqual('my/custom_path/file1.txt', files[0].file_path)
        self.assertEqual('d2c24d65e1d3870f4cf2dbbd8c994b4977a7c384', files[0].file_hash)
        self.assertEqual(21, files[0].size)
        self.assertEqual(
            f"file://{(self.test_dir.resolve() / 'data/file1.txt').as_posix()}",
            files[0].metadata['file_path'])

        self.assertEqual('my/custom_path/hardlinked_file.txt', files[1].file_path)
        self.assertEqual('da3e6ddfa171e1ab5564609caa1dbbea9871886e', files[1].file_hash)
        self.assertEqual(44, files[1].size)
        self.assertEqual(f"file://{(self.test_dir.resolve() / 'data/hardlinked_file.txt').as_posix()}",
                         files[1].metadata['file_path'])

        self.assertEqual('my/custom_path/sub_dir/file_in_subdir.txt', files[2].file_path)
        self.assertEqual('98181b1a4c880a462fcfa96b92c84b8e945ac335', files[2].file_hash)
        self.assertEqual(24, files[2].size)
        self.assertEqual(f"file://{(self.test_dir.resolve() / 'data/sub_dir/file_in_subdir.txt').as_posix()}",
                         files[2].metadata['file_path'])

        self.assertEqual('my/custom_path/symlinked_file.txt', files[3].file_path)
        self.assertEqual('da3e6ddfa171e1ab5564609caa1dbbea9871886e', files[3].file_hash)
        self.assertEqual(44, files[3].size)
        self.assertEqual(f"file://{(self.test_sources_dir.resolve() / 'file_to_link.txt').as_posix()}",
                         files[3].metadata['file_path'])

    def test_expand_user(self):
        os.environ['HOME'] = str(self.test_dir.resolve())

        with open(pathlib.Path('~/tmp_test_expand_user').expanduser(), 'w') as f:
            f.write("File to test ~ resolution")

        files = LocalArtifactDriver.get_tracked_files('~/tmp_test_expand_user')

        self.assertEqual(1, len(files))
        file = files[0]

        self.assertEqual('tmp_test_expand_user', file.file_path)
        self.assertEqual('eb596bf2f5fd0461d3d0b432f805b3984786c721', file.file_hash)
        self.assertEqual(25, file.size)

    def test_wildcards_not_supported(self):
        with self.assertRaises(NeptuneUnsupportedArtifactFunctionalityException):
            LocalArtifactDriver.get_tracked_files(str(self.test_dir / 'data/*.txt'))



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import os
import stat
import tarfile

from future.builtins import object

from neptune.internal.hardware.constants import BYTES_IN_ONE_MB


class FileChunk(object):
    def __init__(self, data, start, end):
        self.data = data
        self.start = start
        self.end = end

    def get_data(self):
        return io.BytesIO(self.data)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FileChunkStream(object):

    def __init__(self, upload_entry):
        self.filename = upload_entry.target_path
        if upload_entry.is_stream():
            self.fobj = upload_entry.source_path
            self.length = None
            self.permissions = '----------'
        else:
            self.fobj = io.open(upload_entry.source_path, 'rb')
            self.length = os.path.getsize(upload_entry.source_path)
            self.permissions = self.permissions_to_unix_string(upload_entry.source_path)

    @classmethod
    def permissions_to_unix_string(cls, path):
        st = 0
        if os.path.exists(path):
            st = os.lstat(path).st_mode
        is_dir = 'd' if stat.S_ISDIR(st) else '-'
        dic = {'7': 'rwx', '6': 'rw-', '5': 'r-x', '4': 'r--', '3': '-wx', '2': '-w-', '1': '--x', '0': '---'}
        perm = ("%03o" % st)[-3:]
        return is_dir + ''.join(dic.get(x, x) for x in perm)

    def __eq__(self, fs):
        if isinstance(self, fs.__class__):
            return self.__dict__ == fs.__dict__
        return False

    def generate(self, chunk_size=BYTES_IN_ONE_MB):
        last_offset = 0
        while True:
            chunk = self.fobj.read(chunk_size)
            if chunk:
                if isinstance(chunk, str):
                    chunk = chunk.encode('utf-8')
                new_offset = last_offset + len(chunk)
                yield FileChunk(chunk, last_offset, new_offset)
                last_offset = new_offset
            else:
                if last_offset == 0:
                    yield FileChunk(b'', 0, 0)
                break

    def close(self):
        self.fobj.close()


def compress_to_tar_gz_in_memory(upload_entries):
    f = io.BytesIO(b'')

    with tarfile.TarFile.open(fileobj=f, mode='w|gz', dereference=True) as archive:
        for entry in upload_entries:
            archive.add(name=entry.source_path, arcname=entry.target_path, recursive=True)

    f.seek(0)
    data = f.read()
    return data
